/* this file is a part of amp software, (C) tomislav uzelac 1996,1997
 */

/* this is audio.c mangled into a Qt widget */
/* portions not originally in audio.c are by lvoge@cs.vu.nl */
/* mixer and balancer by andrej.bagon@uni-mb.si */
/* interface worked by andrej.bagon@uni-mb.si */

/* audio.c      main amp source file 
 * Created by: tomislav uzelac  Apr 1996 
 * Karl Anders Oygard added the IRIX code, 10 Mar 1997.
 * Ilkka Karvinen fixed /dev/dsp initialization, 11 Mar 1997.
 * Lutz Vieweg added the HP/UX code, 14 Mar 1997.
 * Dan Nelson added FreeBSD modifications, 23 Mar 1997.
 * Andrew Richards complete reorganisation, new features, 25 Mar 1997
 * Edouard Lafargue added sajber jukebox support, 12 May 1997
 * 
 */
extern "C" {
#include "amp.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#define AUDIO
#include "audio.h"
#include "dump.h"
#include "formats.h"
#include "getbits.h"
#include "huffman.h"
#include "layer2.h"
#include "layer3.h"
#include "position.h"
#include "rtbuf.h"
#include "transform.h"
#include "controldata.h"
#include "guicontrol.h"
}

#include "plDlg.h"
#include <qapp.h>
#include <qfiledlg.h>
#include <qmsgbox.h>
#include <qpixmap.h>
#include <qpopmenu.h>
#include <qstrlist.h>
#include <qtstream.h>
#include <qtimer.h>
#include "QtAMP.h"
#include "pixmaps.h"
#include "backs.h"

#define Inherited QtAMPData

QtAMP::QtAMP (QWidget* parent, const char* name) : Inherited( parent, name ) {/*{{{*/
	setMinimumSize(size());
	setMaximumSize(size());
	setCaption("Xamp");
	
	frameSlider->setStyle(MotifStyle);
	volSlider->setStyle(WindowsStyle);
	balSlider->setStyle(WindowsStyle);

	QSplashBox *splash = new QSplashBox;
	splash->show();
	
	// init mixer
#ifdef OS_Linux
	if ((mixer = open("/dev/mixer", O_RDWR)) < 0)
	    die("Error opening /dev/mixer, exiting\n");
	ioctl(mixer, SOUND_MIXER_READ_PCM, &v);
#endif OS_Linux
	
	/* where mine resource file is located */
	char home[500];
	sprintf(home,"%s",getenv("HOME"));
	sprintf(homeres,"%s/.xamprc",home);
	setenv("HOME",home,1);
	FILE *inFile;
	if (inFile=fopen(homeres,"r"))
	  {
		  char tmp[3];
		  fscanf(inFile,"%s",tmp);
		  tvr=atoi(tmp);
		  fscanf(inFile,"%s",tmp);
		  v=vr=atoi(tmp);
		  fscanf(inFile,"%s",tmp);
		  repeat=atoi(tmp);
		  fscanf(inFile,"%s",tmp);
		  shuffle=atoi(tmp);
	  }
	else
	  {
		  tvr=0;
		  v=vr=v/257;
		  repeat=shuffle=0;
	  }
	
	fclose(inFile);
	/* end of reading the config */

	/* if you want to see what it loads.. go ahead :) */
//	printf("%d\n%d\n%d\n%d\n",tvr,v,repeat,shuffle);
	
	/* end of mixer settings */
	balSlider->setValue( tvr );
	volSlider->setValue( v );
	    
	// put pixmaps on the buttons 
	QPixmap p;
	
/*	p.loadFromData(xamp_pix_xampback_xpm_data,xamp_pix_xampback_xpm_len);*/
	p.loadFromData(backs_xamp33_xpm_data,backs_xamp33_xpm_len);
	setBackgroundPixmap(p);
	
	p.loadFromData(pixmaps_xopy_xpm_data, pixmaps_xopy_xpm_len);
	aboutButton->setPixmap( p );
	p.loadFromData(pixmaps_pause_xpm_data, pixmaps_pause_xpm_len);
	playButton->setPixmap( p );
	p.loadFromData(pixmaps_ff_xpm_data, pixmaps_ff_xpm_len);
	ff_Button->setPixmap( p );
	p.loadFromData(pixmaps_rew_xpm_data, pixmaps_rew_xpm_len);
	rew_Button->setPixmap( p );
	p.loadFromData(pixmaps_x_xpm_data, pixmaps_x_xpm_len);
	x_Button->setPixmap( p );
	p.loadFromData(pixmaps_load_xpm_data, pixmaps_load_xpm_len);
	songButton->setPixmap( p );
	p.loadFromData(pixmaps_stop_xpm_data, pixmaps_stop_xpm_len);
	stop_Button->setPixmap( p );
	p.loadFromData(pixmaps_mstop_xpm_data, pixmaps_mstop_xpm_len);
	whatLabel->setPixmap( p );
	
	if (shuffle)
	    p.loadFromData(shuffle_xpm_data, shuffle_xpm_len);
	else
	    p.loadFromData(shufflen_xpm_data, shufflen_xpm_len);
	shuffleButton->setPixmap( p );
	
	if (repeat)
	    p.loadFromData(pixmaps_repc_xpm_data, pixmaps_repc_xpm_len);
        else
	    p.loadFromData(pixmaps_rep_xpm_data, pixmaps_rep_xpm_len);
	repeatButton->setPixmap( p );
	
	// put arguments in playlist 
	playList = new QStrList;
	playList->setAutoDelete(true);
	argPos = args(qApp->argc(), qApp->argv());
	char s[2550];
	FILE *dat;
	char text[17];
	while (qApp->argv()[argPos])
	  {
		  strcpy(s,qApp->argv()[argPos++]);
		  
//		  printf("Cores before: filename %s\n",s);
		  dat=fopen(s,"r");
		  fgets(text,16,dat);
//		  printf("Cores later or the text is %s\n",text);
		  fclose(dat);
		  if (!strcmp(text,"[Xamp playlist]"))
		    {
			    QFile f(s);
			    QTextStream sf(&f);
			    f.open(IO_ReadOnly);
			    QString t;
			    t = sf.readLine();
			    while (!sf.eof())
			      {
				      t =sf.readLine();
				      if (t.length() > 0)
					  playList->append(t);
			      }
//		    printf("Looks like it is a Xamp playlist\n");
		    }
		  else
		      playList->append(s);
	  }
	playList->first();
	
	// misc setup
	update_label(0);
	slider_being_dragged = false;

	// init playDialog 
	playDialog = new plDlg();
	connect(playDialog, SIGNAL(fetch_list()), SLOT(get_playlist()));
	connect(playDialog, SIGNAL(seek_list()), SLOT(seek_playlist()));
	playDialog->hide();
	playDialog->setList(playList);
	
	/* get the right MONO/STEREO info */
	downmix=A_DOWNMIX;
	a_counter=0;
	strcpy(fil," ");

	/* the about dialog */
	aboutDialog = new about();
	aboutDialog->hide();

	srand((unsigned)time(NULL));
	timer = NULL;
	
	/* if he has shuffle.. lets shuffle the first one too :) */
	if (shuffle) 
	  {
		  int i;
		  do 
		    {
			    i = (int)((double)playList->count() * rand() / (RAND_MAX + 1.0));
		    }
		  while (i == playList->at());
		  playList->at(i);
	  }
	
	play_stream(); // play the first/shuffled song
}

QtAMP::~QtAMP() {
	delete playList;
	delete playDialog;
#ifdef OS_Linux
	close(mixer);
#endif
}

void QtAMP::statusDisplay(void) {
        int seconds10x;
	
	if (!slider_being_dragged && !(cnt % 10)) {
		seconds10x = (cnt * 11520) / 44100;
		/* lodewijk why dont you learn something?! */
		/* if the damn 44100 is set it will play right the 22050 
		 * and all other mp3s.. it is tested ( i have a bunch
		 * of those mp3s so trust me plz :) 
		 */
/*		seconds10x = (cnt * 11520) / t_sampling_frequency[header.ID][header.sampling_frequency]; */
		frameSlider->setValue(seconds10x); 
	}
}

void QtAMP::show_header(void) {
	int bitrate=t_bitrate[header.ID][3-header.layer][header.bitrate_index];
	int fs=t_sampling_frequency[header.ID][header.sampling_frequency];
	int mpg,layer;
	QString s;
        layer=4-header.layer;
        mpg = (header.ID==1) ? 1 : 2;
	s.sprintf("%d", (int) fs / 1000);
	propLabel->setText( s );
	s.sprintf("%d", bitrate);
	bitLabel->setText( s );
	
	s.sprintf("%u",playList->count());
	allLabel->setText( s );
	
	s.sprintf("%u",playList->at()+1);
	curLabel->setText( s );	
	
	QPixmap p;
	p.loadFromData(pixmaps_mplay_xpm_data, pixmaps_mplay_xpm_len);
	whatLabel->setPixmap( p );
	
	if (!downmix && (header.mode != 3))
	    modeButton->setText("STEREO");
	else
	    modeButton->setText("MONO");
}

int QtAMP::numSeconds(void) {
	// FIXME: account for stereo/mono ?
	return (int) (statbuf.st_size / (t_bitrate[header.ID][3-header.layer][header.bitrate_index] * 125));
}

void QtAMP::vol_slider_moved(int vol) 
{
	char volum[15];
	int tmp;
	v=vol;
	int tv=0,tvl=0;
	if (tvr < 0)	
	  {
		  tmp=v;
		  tv=((v*256)+(tvr*256))/256;
		  tvl=tmp-tvr;
	  }
	else
	  {
		  tmp=v;
		  tv=((v*256)+(tvr*256))/256;
		  tvl=tmp-tvr;
		  if (tvl < 0)
		      tvl=0;
	  }
	vr=vol;
	sprintf(volum,"Volume: %d",vr);
	playLabel->setText( volum );
	//	printf("%d %d\n",tvr,tv);
#ifdef OS_Linux
	int x = (tv*256 + tvl);
	ioctl(mixer, SOUND_MIXER_WRITE_PCM, &x);
#endif
}

void QtAMP::bal_slider_moved(int bal)
{
	char volum[15];
	tvr=bal;
	sprintf(volum,"Balance: %d",tvr);
	vol_slider_moved(vr);
	playLabel->setText( volum );
}

void QtAMP::bal_slider_released()
{
	if (timer)
	    playLabel->setText(fil);
	else
	    playLabel->setText("No Audio");
}

void QtAMP::vol_slider_released()
{
	if (timer)
	    playLabel->setText(fil);
	else
	    playLabel->setText("No Audio");
}

void QtAMP::song_clicked() {
	QString s;
	playDialog->show();
	s.sprintf("%u",playList->count());
	allLabel->setText( s );
	s.sprintf("%u",playList->at()+1);
	curLabel->setText( s );
}

void QtAMP::get_playlist() {
	int i = playList->at();
	playDialog->getList(playList);
	if (playList->count() == 0)
	  stop_stream();
	else if ((i != -1) && !playList->at(i))
	  playList->last();
}

void QtAMP::seek_playlist() {
	stop_stream();
	playDialog->getList(playList);
	play_stream();
}

// playing routines 
void QtAMP::decodeMPEG(const char *inFileStr) {
	const char *p;
	
        if (strcmp(inFileStr, "-") == 0)
	  in_file = stdin;
        else {
                if ((in_file = fopen(inFileStr, "r")) == NULL) {
                        warn("Could not open file: %s\n", inFileStr);
                        return;
                }
		stat(inFileStr, &statbuf);
        }
	
	p = inFileStr;
	while (*++p) ;
	while ((p > inFileStr) && (*--p != '/')) ;
	if (*p == '/') p++;
	sprintf(fil,"%d. %s",playList->at()+1,p);
	int xa;
	for (xa=1;xa<strlen(fil);xa++)
	  {
		  if ( fil[xa] == '_')
		      fil[xa]=' ';     /* was fil */
	  }
	playLabel->setLabel(fil); /* was p */
	setCaption(QString("Xamp: ") + QString(fil));

	/* this reads the first header, opens the audio device with the
	 * relevant values, seeks back, plays the frame to record the
	 * frame_size and plays the second frame to see if the mp3 is padded
	 * or not */
	
	cnt = 0;
	read_header();
	A_DOWNMIX=downmix;
	
	if (AUDIO_BUFFER_SIZE == 0)
	  audioOpen(t_sampling_frequency[header.ID][header.sampling_frequency],
		    (!A_DOWNMIX && (header.mode != 3)), A_SET_VOLUME);
	else
	  audioBufferOpen(t_sampling_frequency[header.ID][header.sampling_frequency],
			  (!A_DOWNMIX && (header.mode != 3)), A_SET_VOLUME);
	fseek(in_file, 0, SEEK_SET);
	play_frame();
	frame_size = ftell(in_file) + 1;
	play_frame();
	padded = ftell(in_file) == (2 * frame_size - 1);
	if (!padded)
	  --frame_size;
//	printf("frame_size: %d, %s\n", frame_size, (padded) ? "padded" : "not padded");
	bitindex = header.bitrate_index;
	may_seek = true;
	if ((header.mode == 2) &&
	    (t_bitrate[header.ID][3-header.layer][bitindex] != 256))
	  frame_size--;
	frameSlider->setRange(0, 10 * numSeconds());
	frameSlider->setTickInterval((10 * numSeconds()) / 50);
	show_header();

        append = data = nch = 0;        /* initialize globals */
        f_bdirty = TRUE;
        bclean_bytes = 0;
	
//	repeat = false; /* if the guy wanna repeat let him repeat right? */
	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), SLOT(play_frame()));
	timer->start(0);
}

bool QtAMP::read_header(void) {
	int g;

	if ((g = gethdr(&header)) != 0) {
		switch (g) {
		 case GETHDR_ERR:
			die("error reading mpeg bitstream. exiting.\n");
			break;
		 case GETHDR_NS:
			warn("this is a file in MPEG 2.5 format, which is not defined\n");
			warn("by ISO/MPEG. It is \"a special Fraunhofer format\".\n");
			warn("amp does not support this format. sorry.\n");
			break;
		 case GETHDR_FL1:
			warn("ISO/MPEG layer 1 is not supported by amp (yet).\n");
			break;
		 case GETHDR_FF:
			warn("free format bitstreams are not supported. sorry.\n");
			break;
		 case GETHDR_SYN:
			warn("oops, we're out of sync.\n");
			break;
		 case GETHDR_EOF:
			if (A_FORMAT_WAVE)
			  wav_end(&header);
			break;
		 default:
			break;
		}
		if (repeat)
		  seek_to_frame(0);
		else
		  {
			  QString s;
			  next_clicked();
			  s.sprintf("%u",playList->at()+1);
			  curLabel->setText( s );
		  }
		return false;
	}
	return true;
}
	
void QtAMP::play_frame(void) {

	if (!read_header())
	  return;
	
#ifdef ANAL
	if (header.bitrate_index != bitindex) {
		may_seek = false;
		QMessageBox box;
		box.setText("This stream has a variable bitrate\n"
			    "Seeking with the slider is disabled\n"
			    "for this stream.");
		box.exec();
	}
#endif
	
	/* crc is read just to get out of the way.
	 */
	if (header.protection_bit == 0)
	  getcrc();
	
	statusDisplay();
	
	if (header.layer == 1) {
		if (layer3_frame(&header, cnt)) {
			warn(" error. blip.\n");
			stop_stream();
			return;
		}
	} else if (header.layer == 2)
		  if (layer2_frame(&header, cnt)) {
			  warn(" error. blip.\n");
			  stop_stream();
			  return;
		  }
	++cnt;
}

void QtAMP::stop_stream() {
	if (!timer)
	  return;
	delete timer;
	timer = NULL;
	may_seek = false;
	fclose(in_file);
	if (AUDIO_BUFFER_SIZE != 0)
	  audioBufferClose();
	else
	  audioClose();
	if (paused) {
		paused = false;
		QPixmap p;
		
		p.loadFromData(pixmaps_pause_xpm_data, pixmaps_pause_xpm_len);
		playButton->setPixmap(p);
	}		  
}

void QtAMP::seek_to_frame(int frame) {
	/* I don't know the rationale behind these numbers.
	 * it's extremely likely this will fail for weird bitrates and modes.
	 * I've only tested what I could encode here and what I have on cd.
	 */
	/* 11 Aug 97, I just got a boatload of 128kbps 44kHz (HIP97 :) that for
	 * some reason _don't_ have any short frames. fortunately, all these
	 * files have a framesize of 417 so it's easy to catch
	 */
	int ofs = frame_size * frame - (frame > 0);
	if (padded) {
		switch(t_bitrate[header.ID][3-header.layer][header.bitrate_index]) {
		 case 56 : ofs -= frame / 7; break;
		 case 64 : ofs -= frame / 49; break;
		 case 96 : ofs -= 9 * (frame / 17) + (frame % 17 == 0) + frame % 17 / 2; break;
		 case 112 : ofs -= 2 * (frame / 7) + (frame % 7 >= 4); break;
		 case 128 : ofs -= 2 * (frame / 49) + (frame % 49 >= 25); break;
		 case 256 : ofs -= 4 * (frame / 49) + (frame % 49 >= 13) +
			  ((frame % 49 > 13) ? ((frame % 49) - 13) / 12 : 0); break;
		 default: die("magic values not known for this bitrate, please mail lvoge@cs.vu.nl");
		}
	}
	fseek(in_file, ofs, SEEK_SET);
	cnt = frame;
}


// progress-slider handling routines 
// these two are necessary to keep the slider from being updated while the
// slider is being slided.
void QtAMP::slider_pressed() {
	slider_being_dragged = true;
}

void QtAMP::slider_released() 
{
	slider_being_dragged = false;
	if (timer)
	  {
		  int frame = (sec10x * 44100 / 10) / 1152;
		  if (may_seek)
		      seek_to_frame(frame);
	  }
}


void QtAMP::slider_moved(int sec10)
{
	sec10x=sec10;
}

void QtAMP::update_label(int sec) {
	QString s;
	sec /= 10;
	s.sprintf("%02d:%02d", (int) sec / 60, (int) sec % 60);
	frameLabel->setText( s );
}

// functions handling the buttons 
void QtAMP::rewind_clicked() {
	stop_stream();
	playList->first();
	play_stream();
}

void QtAMP::prev_clicked(void) {
	stop_stream();
	if (playList->at() > 0)
	  playList->prev();
	else
	  playList->last();
	QString s;
	s.sprintf("%u",playList->at()+1);
	curLabel->setText( s );
	play_stream();
}

void QtAMP::play_pressed() {
	QPixmap p;
	if (timer) {
		if (paused) {
			timer->start(0);
			p.loadFromData(pixmaps_mplay_xpm_data,pixmaps_mplay_xpm_len);
			whatLabel->setPixmap(p);
			p.loadFromData(pixmaps_pause_xpm_data, pixmaps_pause_xpm_len);
		} else {
			timer->stop();
			p.loadFromData(pixmaps_mpause_xpm_data,pixmaps_mpause_xpm_len);
			whatLabel->setPixmap(p);
			p.loadFromData(pixmaps_play_xpm_data, pixmaps_play_xpm_len);
		}
		playButton->setPixmap(p);
		paused = !paused;
	} 
	else
	  {
		  play_stream();
		  p.loadFromData(pixmaps_pause_xpm_data, pixmaps_pause_xpm_len);
		  playButton->setPixmap(p);
	  }		  
}

void QtAMP::play_stream()
{
	paused = 0;
	if (playList->current())
	  decodeMPEG(playList->current());
}

void QtAMP::next_clicked()
{
	stop_stream();
	if (shuffle)
	  {
		  int i;
		  do
		    {
			    i = (int)((double)playList->count() * rand() / (RAND_MAX + 1.0));
		    } 
		  while (i == playList->at());
		  playList->at(i);
	  }
	else
	  {
		  playList->next();
		  if (playList->at() == -1)
		      playList->first();   /* wrap */
	  }
	if (playList->at() != -1)
	    play_stream();
	
	QString s;
	s.sprintf("%u",playList->at()+1);
	curLabel->setText( s );
}

void QtAMP::ff_clicked() {
	stop_stream();
	playList->last();
	play_stream();
}

void QtAMP::repeat_toggled()
{
	repeat = !repeat;
	QPixmap pm;
	if (repeat)
	    pm.loadFromData(pixmaps_repc_xpm_data, pixmaps_repc_xpm_len);
	else
	    pm.loadFromData(pixmaps_rep_xpm_data, pixmaps_rep_xpm_len);
	repeatButton->setPixmap( pm );
}

void QtAMP::shuffle_toggled() {
	QPixmap pm;
	if (shuffle)
	    pm.loadFromData(shufflen_xpm_data, shufflen_xpm_len);
	else
	    pm.loadFromData(shuffle_xpm_data, shuffle_xpm_len);
	
	shuffleButton->setPixmap( pm );
	shuffle = !shuffle;
}

/*
void QtAMP::about_dialog(void)
{
}
*/

QSplashBox::QSplashBox(void) : QMessageBox(0, 0) {
	setText("Xamp 0.7.6, interface Andrej Bagon 1997\n"
		"frontend for Xamp (C) Lodewijk Voge 1997\n"
		"based on amp 0.7.6, (C) Tomislav Uzelac 1996,1997\n\n"
		"THIS PROGRAM COMES WITH ABSOLUTELY NO WARRANTY\n"
		"PLEASE READ THE DOCUMENTATION FOR DETAILS");
	QTimer::singleShot((int)(2*1000), this, SLOT(done()));
}

void QSplashBox::done()
{
	delete this;		       /* self expression */
}

/* xopy "add ons" */
void QtAMP::real_quit()
{
	FILE *outFile;
	int dolzina;
	
	outFile=fopen(homeres, "w");
	
	/* here are the resources i save */
	fprintf(outFile,"%d\n%d\n%d\n%d\n",tvr,v,repeat,shuffle);

	fclose(outFile);
	if (timer)
	    stop_stream();
	exit(0);
}

void QtAMP::show_about()
{
	aboutDialog->show();
}

void QtAMP::mode_clicked()
{
	downmix=!downmix;
	/* i dont know how to set the dsp directly to mono :( */
	/* so i need to stop it and then to start it from the begining */
	stop_stream();
	if (!paused)
	  {
		  QPixmap p;
		  p.loadFromData(pixmaps_mplay_xpm_data,pixmaps_mplay_xpm_len);
		  whatLabel->setPixmap( p );
		  p.loadFromData(pixmaps_pause_xpm_data, pixmaps_pause_xpm_len);
		  playButton->setPixmap( p );
	  }
	play_stream();
}

void QtAMP::real_stop()
{
	if (timer)
	  {
		  stop_stream();
		  QPixmap p;
		  p.loadFromData(pixmaps_mstop_xpm_data, pixmaps_mstop_xpm_len);
		  whatLabel->setPixmap( p );
		  p.loadFromData(pixmaps_play_xpm_data, pixmaps_play_xpm_len);
		  playButton->setPixmap( p );
	  }
	stop_stream();
	timer=0;
	paused=0;
}

