/* Copyright 1997, lvoge@cs.vu.nl */
#ifndef QTAMP_H
#define QTAMP_H

#include <sys/stat.h>
#include "QtAMPData.h"
#include "audio.h"
#include "about.h"

#include <qmsgbox.h>

class plDlg;
class QStrList;
class QTimer;

class QtAMP : public QtAMPData {
	Q_OBJECT
 public:
	QtAMP (QWidget* parent = NULL, const char* name = NULL);
	virtual ~QtAMP(void);
 private slots:
	void slider_pressed();	       /* suspends slider & label updates */
	void slider_released();	       /* resumes updates */
	void slider_moved(int);	       /* calls seek_to_frame */
	void play_pressed();	       /* pause or play */
	void play_frame();	       /* play a single frame */
	void prev_clicked();
	void next_clicked();
	void vol_slider_moved(int);
	void bal_slider_moved(int);
	void vol_slider_released();
	void bal_slider_released();
	void song_clicked();
	void rewind_clicked();
	void ff_clicked();
	void repeat_toggled();
	void shuffle_toggled();
/*	void about_dialog(); */
	void get_playlist();	       /* fetch playlist without seeking */
	void seek_playlist();	       /* fetch and seek */
	void real_quit();		/* clear everytihng & quit */
	void real_stop();		/* stop the song */
	void show_about();		/* the about dialog */
	void mode_clicked();		/* MONO <-> STEREO downmix */
	
 private:
	int argPos,		       /* position in the playlist */
	  cnt,			       /* number of current frame */
	  frame_size,		       /* size of the first frame */
	  mixer,		       /* handle for /dev/mixer */
	  bitindex;		       /* bitrate of first frame of stream */
	bool slider_being_dragged, paused, repeat, may_seek, padded, shuffle;
	struct stat statbuf;	       /* a struct stat of the current file */
	struct AUDIO_HEADER header;    /* header of the current frame */
	QTimer *timer;		       /* dummy timer, empties the Qt queue
					* and calls play_frame */
	QStrList *playList;	       /* list of filenames to be played */
	plDlg *playDialog;
	about *aboutDialog;
	int downmix,a_counter,v,tvr,vr; /* sorry but cant make it diferent */
	char homeres[200];              /* we will have the home resource */
	char fil[2500];
	int sec10x;

	void decodeMPEG(const char *); /* start playing file */
	int numSeconds(void);	       /* approximates the length of the
					* current song */
	void play_stream(void);	       /* play *playList->current() */
	bool read_header(void);	       /* read header from file */
	void seek_to_frame(int);       /* seek to specified frame */
	void show_header(void);	       /* per-file GUI filling */
	void statusDisplay(void);      /* per-frame GUI updating */
	void stop_stream(void);	       /* stop playing the current stream */
	void update_label(int);	       /* puts the label above the slider */
};

/* suicidal message box */
class QSplashBox : public QMessageBox {
	Q_OBJECT
 public:
	QSplashBox(void);
 public slots:
	void done();
};

#endif /* QTAMP_H */
